"""
A set of fast glyph and font objects for math operations.

This was inspired, and is more or less a clone, of Erik van Blokland's
brilliant glyph math in RoboFab.
"""

from fontMath.mathGlyph import MathGlyph
from fontMath.mathInfo import MathInfo
from fontMath.mathKerning import MathKerning


version = __version__ = "0.6.0"
